; SW pro registraci tel. karet
; Ing. Petr Novak (061) 1.8.1999

; obdobne clanky:
; Amaterske Radio 9/95 str. 23
; PE 9/98 str. 28
; A Radio 6/99 str. 29

; data v tel. karte (prvnich 96 bitu, ulozeno v R0...R)
; (!!! nizsi bit na vyssi hodnotu !!!)
; proto jsou bity nacteny v opacnem poradi do bytu
; (bit   7  6  5  4  3  2  1  0)
; R0  - 00 01 02 03 04 05 06 07 -> kontrolni soucet (216 - pocet jednicek)
; R1  - 08 09 10 11 12 13 14 15 -> 83h (pro tel. karty)
; R2  - 16 17 18 19 20 21 22 23 -> 1Xh (pro tel. karty)
; R3  - 24 25 26 27 28 29 30 31 -> XXh (XXXh pocet jednotek na nove karte v BDC)
; R4  - 32 33 34 35 36 37 38 39 -> kod vyrobce karty
; R5  - 40 41 42 43 44 45 46 47 -> seriove cislo (cast 5)
; R6  - 48 49 50 51 52 53 54 55 -> seriove cislo (cast 4)
; R7  - 56 57 58 59 60 61 62 63 -> seriove cislo (cast 3)
; R8  - 64 65 66 67 68 69 70 71 -> seriove cislo (cast 2)
; R9  - 72 73 74 75 76 77 78 79 -> seriove cislo (cast 1)
; R10 - 80 81 82 83 84 85 86 87 -> 11h (pro let. karty)
; R11 - 88 89 90 91 92 93 94 95 -> 55h (kod zeme)

; pro jednoduchost se do EEPROM ukladaji pouze hodnoty v R5, R6, R7, R8, R9

; zapojeni pinu
; PB3 - OUT (vystup) =1 pro platnou kartu
; PB4 - vstup rozpinaciho kontaktu, vystup pro RESET karty a vystup pro LED_ERROR
; PB2 - vstup pro tlacitko a vystup pro LED_OK
; PB1 - vystup pro CLK karty
; PB0 - vstup ctenych dat z karty

; obsazeni registru
; R0 az R11 -> prectena karta

; R16,R17,R18,R19 vseucelove

; R20 -> ukazatel na prvni volnou pozici v EEPROM pro zapis serioveho cisla
;	 nebo na prave nalezene seriove cislo

; R22 -> rychlost blikani LED_ERROR
;	(pomale -> karta je platna, ale neni zaevidovana)
;	(rychle -> karta je neplatna)

; R26,R27 -> pro cekaci smycku
; R28 -> bitove priznaky
;   R28.1=1 -> souhlasi kontrolni soucet karty
;   R28.2=1 -> seriove cislo bylo v EEPROM nalezeno (jeho zacatek je v R20)
; R31:R30 -> Z pro zapis/cteni hodnot

.include "2343def.inc"		; pouzity procesor AT90S2343

.equ	FIRST_SER	= 10	; adresa (zacatku) prvniho serioveho cisla v EEPROM
.equ	MAX_SER		= 10	; celkovy mozny pocet ulozenych seriovych cisel

.equ	FLASH_SLOW	= 255	; pomale blikani LED_ERR
.equ	FLASH_FAST	= 100	; rychle blikani LED_ERR

.equ	LED_OK		= 2	; pripojeni LED_OK (PB2)
.equ	OUT		= 3	; vystupni signal (1 -> spravna karta)
.equ	LED_ERR		= 4	; pripojeni LED_ERR (PB4)

.equ	DATA		= 00h	; ulozeni nactenych dat

		.cseg			; programovy segment
		.org 	0		; zacatek programove pameti
		rjmp RESET		; skok do programu

		.org 	3		; zde zacina skutecny program
RESET:		
		ldi	R16,$D0		; inicializace zasobniku
		out	SPL,R16		; pomoci R16 na hodnotu DFh (konec SRAM)

		; nastaveni vstupu a vystupu
		ldi	R16,0b00001110	; PB3, PB2 a PB1 jsou vystupni
		out	DDRB,R16
		sbi	PORTB,4		; PULL-UP na vstupu

		; pocatecni nastaveni
		ldi	R31,00		; data ulozena do adresy 255

NEW_CARD:	sbi	PORTB,LED_OK	; LED_OK nesviti
		cbi	PORTB,OUT	; vystup neaktivni
NEW_CARD_01:	ldi	R22,FLASH_FAST	; karta je neplatna
NEW_CARD_02:	; cekani na vytazeni karty
		rcall	DELAY		; cekej
		sbis	PINB,4		; pokud je karta (1) -> preskok
		rjmp	WAIT_CARD	; karta byla vytazena -> cekej na novou
		; pokud je zasunuta neplatna karta blika se LED_ERROR
		sbi	DDRB,4		; PB4 na vystup
		cbi	PORTB,LED_ERR	; LED_ERR sviti
		mov	R27,R22		; cekej
		rcall	DELAY01
		sbi	PORTB,LED_ERR	; LDE_ERR nesviti
		mov	R27,R22
		rcall	DELAY01		; jekej
		cbi	DDRB,4		; PB4 na vstup
		rjmp	NEW_CARD_02	; cekej dal na vytazeni karty

WAIT_CARD:	; cekani na vlozeni nove karty
		sbis	PINB,4		; pokud PB4=1 -> preskok
		rjmp	WAIT_CARD

		rcall	READ_CARD	; precteni karty

		rcall	CHECK_CARD	; kontrola platnosti dat

		; test zda je vlozena platna karta (nemusi byt jeste spravna)
		bst	R28,1		; R28.1 -> T
		brtc	NEW_CARD	; pokud T=0 (neplatna karta) -> cekej na novou

		; je skutecne vlozena platna (jeste ne spravna) karta
		ldi	R22,FLASH_SLOW	; karta je aspon platna

		rcall	SEARCH_SER	; nalezeni zda je ser. cislo zname
					; pokud R28.2=1 -> karta je spravna
					; R20 udava adresu v EEPROM ulozeni ser. cisla

		; test na stisk tlacitka pro odevidovani/prievidovani ser. cisla
		cbi	DDRB,2		; PB2 na vstup
		sbi	PORTB,2		; zapojit PULL-UP
		rcall	DELAY		; cekej
		sbis	PINB,2		; pokud tlacitko nezmacknuto -> preskok
		rjmp	CHANGE		; jinak je pozadovan zapis/smazani ser. cisla

		; test zda je vlozena spravna karta
		bst	R28,2		; R28.2 -> T
		brtc	NEW_CARD_02	; pokud T=0 (nespravna karta) -> cekej na novou

		; je vlozena spravna karta
		sbi	DDRB,2		; PB2 na vystup
		cbi	PORTB,LED_OK	; rosvit LED_OK
		sbi	PORTB,OUT	; vystup aktivni

		; cekani na vytazeni spravne karty nebo uplynuti asi 10s kdy je karta
		; opet otestovana
		cbi	DDRB,4		; PB4 na vstup (bude se cist)
		ldi	R16,40		; cela smycka asi 10s
SPRAVNE_01:	ldi	R17,255
SPRAVNE_02:	rcall	DELAY		; cekej
		sbis	PINB,4		; pokud karta vlozena (1) -> preskok
		rjmp	NEW_CARD	; jinak cekej na vlozeni nove kraty
		dec	R17
		brne	SPRAVNE_02
		dec	R16
		brne	SPRAVNE_01
		rjmp	WAIT_CARD	; po vyprseni tohoto case -> opet test

		; obsluzeni smazani/pridani ser. cisla
		; cekani na uvolneni tlacitka
CHANGE:		sbis	PINB,2		; pokud tlacitko uvolneno -> preskok
		rjmp	CHANGE		; jinak cekej na jeho uvolneni
		ldi	R27,255		; cekej za zakmity
		rcall	DELAY01
		bst	R28,2		; 0 -> nezname / 1-> zname (ser. cislo)
		brtc	ADDING		; pokud T=1 je zname -> pozadavek smazani
		; smazani znameho ser. cisla
		rcall	ERASE_SER	; smaz ser. cislo v EEPROM na pozici (R20)
		rjmp	WAIT_CARD	; znova do testu
ADDING:		; pridani noveho ser. cisla
		rcall	FREE_SER	; nalezeni volneho mista (vraceno v R20)
		bst	R28,2		; R28.2 -> T 
		brtc	NEW_CARD	; pokud T=0 (misto nenalezeno) -> chyba
		rcall	WRITE_SER	; zapis noveho ser. cisla (na pozici R20)
		rjmp	WAIT_CARD	; znova do testu
;-------------------------------------------------------------------------------
;             ### podprogram pro cteni obsahu karty ###
; vystup: prectena karta v registrech R0 az R11

READ_CARD:	ldi	R30,DATA	; misto pro ukladani ctenych dat

		; generovani RESETu pro kartu
		sbi	DDRB,4		; PB4 na vystup
		cbi	PORTB,4		; PB4=0 (aktivni RESET)
		rcall	DELAY		; cekej

		sbi	PORTB,1		; CLK=1
		rcall	DELAY		; cekej
		cbi	PORTB,1		; CLK=0

		rcall	DELAY		; cekej
		sbi	PORTB,4		; RES=1 (neaktivni RESET)
		rcall	DELAY		; cekej
		
		; vycteni vsech bitu
		ldi	R17,12		; bude 12 bytu
READ_CARD_01:	; vycteni 8 bitu
		clr	R15		; 00h -> R15
		ldi	R16,8		; bude se cist 8 bitu
READ_CARD_02:	clc			; 0 -> Carry
		sbic	PINB,0		; pokud PB0=0 -> preskoc
		sec			; 1 -> Carry
		rol	R15		; rotace o bit <<  (Carry -> bit0) otoceni poradi bitu

		; hodinovy puls
		sbi	PORTB,1		; CLK=1
		rcall	DELAY		; cekej
		cbi	PORTB,1		; CLK=0
		rcall	DELAY		; cekej

		dec	R16		; jeden bit precten
		brne	READ_CARD_02	; precti vsechny bity
		st	Z+,R15		; uloz R15 a posun Z o jednu nahoru
		dec	R17		; dalsi osmice prectena
		brne	READ_CARD_01	; precti vsechny osmice

		ret			; navrat z podporgramu
;-------------------------------------------------------------------------------
;               ### podprogram pro kontrolu dat karty ###
; vstup:  prevtana krta v registrech R0 az R11
; vystup: R28.2 =0 -> chybny kontrolni soucet nebo spatna data
;	  R28.2 =1 -> spravny kontrolni soucet a dobra data

CHECK_CARD:	; nastaveni ukazatele pro cteni hodnot
		ldi	R30,DATA+1	; prvni byte se nekontroluje

		; odecteni poctu jednickovych bitu od cisla 216
		ldi	R18,216		; od teto hodnoty se bity odecitaji
		ldi	R17,11		; bude 11 bytu
CHECK_CARD_01:	; odecitani vzdy po 8-mi bitech
		ldi	R16,8		; bude 8 bitu v bytu
		ld	R15,Z+		; hodnota z (Z) do R15 a posun Z o jednicku
CHECK_CARD_02:	ror	R15		; rotace >> -> Carry
		sbci	R18,00		; odecti Carry od R18
		dec	R16		; jeden bit precten
		brne	CHECK_CARD_02	; vsechny bity
		dec	R17		; dalsi osmice prectena
		brne	CHECK_CARD_01	; precti vsechny osmice

		; kontrola spravnosti
		andi	R28,0b11111101	; predpokladej chybu
		; kontrola kontrolniho souctu
		cp	R0,R18		; porovnej R0==R18
		brne	CHECK_END	; pokud R0<>R18 -> chyba, konec
		; kontrola karty pro telefon 831h
		mov	R18,R1		; R1 -> R18
		cpi	R18,$83
		brne	CHECK_END	; pokud R1<>83h -> chyba, konec
		mov	R18,R2		; R2 -> R18
		andi	R18,$f0		; pouze bity 7,6,5,4
		cpi	R18,$10
		brne	CHECK_END	; pokud R2(7,6,5,4]<>1h -> chyba, konec
		; kontrola cisla 11h
		mov	R18,R10		; R10 -> R18
		cpi	R18,$11
		brne	CHECK_END	; pokud R10<>11h -> chyba, konec
		; kontrola kodu zeme (55h)
		mov	R18,R11		; R10 -> R18
		cpi	R18,$55
		brne	CHECK_END	; pokud R11<>55h -> chyba, konec
		
		; vse probehlo kontrolou dobre
		ori	R28,0b00000010	; vse je v poradku
CHECK_END:	ret			; navrat z podporgramu
;-------------------------------------------------------------------------------
;       ### podprogram pro prohledani EEPROM zda toto cislo existuje ###
; vstup:  R5,R6,R7,R8,R9 -> hledane seriove cislo
; vystup: R20 -> ukazaje na nalezene seriove cislo do EEPROM
;	  R28.2=1 -> seriove cislo bylo skutecne nalezeno

SEARCH_SER:	ldi	R25,DATA+5	; zacatek ser. cisla na DATA+5

SEARCH_SER_01:	ldi	R20,FIRST_SER	; pozice ulozeni prvniho ser. cisla
		ldi	R19,MAX_SER	; maximalni pocer ulozenych ser. cisel

SEARCH_SER_02:	mov	R30,R25		; zacatek hledanych dat
		ldi	R18,5		; seriove cislo ma 5 bytu

		ori	R28,0b00000100	; predpokladej shodu
SEARCH_SER_03:	out	EEAR,R20	; adresa pro cteni
		inc	R20		; posun ukazatele v EEPROM
		sbi	EECR,EERE	; start cteni
		in	R16,EEDR	; vyzvednuta data z EEPROM -> R15
		ld	R17,Z+		; vyzvednuta data z registru -> R14
		cpse	R16,R17		; pokud R14==R15 -> preskok
		andi	R28,0b11111011	; neni shoda
		dec	R18		; jeden byte ser. cisla otestovan
		brne	SEARCH_SER_03	; otestuj vsech pet bytu

		; test zda nebylo ser. cislo jiz nalezeno
		bst	R28,2		; R28.2 -> T (1 -> ser. cislo nalezeno)
		brtc	SEARCH_SER_04	; pokud jeste nebylo nalezeno -> skok
		subi	R20,5		; nastav ukazatel na zacatek ser. cisla
		ret			; navrat z podprogramu

SEARCH_SER_04:	dec	R19		; jedno cele cislo otestovano
		brne	SEARCH_SER_02	; otestuj vsechny pokud je treba
		ret			; navrat z podprogramu
;-------------------------------------------------------------------------------
;  ### nalezeni prvni volne pozice v EEPROM pro zapis noveho ser. cisla ###
; vystup: R20 -> udava adresu pro zapis noveho ser. cisla

FREE_SER:	ldi	R16,255		; pro umisteni noveho serioveho cisla
		mov	R10,R16		; se hleda pet cistych adres v EEPROM
		mov	R11,R16		;
		mov	R12,R16
		mov	R13,R16
		mov	R14,R16
		ldi	R25,10		; hodnoty k porovnani jsou od adresy 10
		rjmp	SEARCH_SER_01	; dale jako hledani ser. cisla
;-------------------------------------------------------------------------------
;               ### pro zapis/smazani serioveho cisla do/z EEPROM ###
; vstup:  R5,R6,R7,R8,R9 -> seriove cislo pri zapisu
;	  R20 -> adresa kam se seriove cislo zapise / ktere se bude mazat
; vystup: zapsane/smazane seriove cislo

ERASE_SER:	; vstup pro smazani serioveho cisla
		ldi	R16,255		; nuluj seriove cislo
		mov	R5,R16		; jeho prepisem v registrech
		mov	R6,R16		; a pak toto zapis do EEPROM
		mov	R7,R16
		mov	R8,R16
		mov	R9,R16
WRITE_SER:	; vstup pro zapis serioveho cisla
		; nastaveni ukazatele na data
		ldi	R30,DATA+5	; zacatek ser. cisla na zacatek+5
		ldi	R16,5		; bude se zapisovat 5 registru
		; zapis jednoho registru
WRITE_SER_01:	ld	R17,Z+		; vyzvedni hodnotu do R17 a posun Z o jednu
		out	EEAR,R20	; adresa ulozeni
		out	EEDR,R17	; data pro ulozeni
		sbi	EECR,EEMWE	; start zapisu
		sbi	EECR,EEWE
WRITE_SER_02:	sbic	EECR,EEWE	; pokud konec zapisu -> preskok
		rjmp	WRITE_SER_02	; jinak cekej na ukonceni zapisu
		inc	R20		; posun ukazatel zapisu
		dec	R16		; jedna hodnota zapsana
		brne	WRITE_SER_01	; zapis vsech peti hodnot
		ret			; navrat z podprogramu
;-------------------------------------------------------------------------------
;                      ### podprogram pro cekani ###
; pouziva R26,R27 pro zpozdeni
; vstup v DELAY01 s naplnenym R27 v poctu asi 0.5ms

DELAY:		ldi	R27,2		; asi 1ms
DELAY01:	ldi	R26,255		; asi 0.5ms
DELAY02:	dec	R26
		brne	DELAY02
		dec	R27
		brne	DELAY01
		ret			; navrat z podprogramu
;--------------------------------------------------------------------------------